@@ -48,18 +48,15 @@ OP_HCONTAINS_KEY = 123
 
 class PostgresqlExtCompiler(QueryCompiler):
     def parse_create_index(self, model_class, fields, unique=False):
-        parts = super(PostgresqlExtDatabase, self).parse_create_index(
+        parts = super(PostgresqlExtCompiler, self).parse_create_index(
             model_class, fields, unique)
-        if any(lambda f: isinstance(f, HStore), fields):
+        if any(map(lambda f: isinstance(f, HStore), fields)):
             parts.insert(-1, 'USING GIST')
         return parts
 
 
 class PostgresqlExtDatabase(PostgresqlDatabase):
     compiler_class = PostgresqlExtCompiler
-    expr_overrides = dict_update(PostgresqlDatabase.expr_overrides, {
-        OP_HUPDATE: '||',
-    })
     field_overrides = dict_update(PostgresqlDatabase.field_overrides, {
         'hash': 'hstore',
     })
@@ -67,6 +64,7 @@ class PostgresqlExtDatabase(PostgresqlDatabase):
         OP_HCONTAINS_DICT: '@>',
         OP_HCONTAINS_KEYS: '?&',
         OP_HCONTAINS_KEY: '?',
+        OP_HUPDATE: '||',
     })
 
     def _connect(self, database, **kwargs):
